// Copyright (c) 2022-2023 Yuki Kishimoto
// Copyright (c) 2023-2024 Rust Nostr Developers
// Distributed under the MIT software license

#[cfg(not(target_arch = "wasm32"))]
use std::net::SocketAddr;
use std::sync::atomic::{AtomicBool, AtomicU64, Ordering};
use std::sync::Arc;
use std::time::Duration;

use crate::client::options::DEFAULT_SEND_TIMEOUT;

pub const DEFAULT_RETRY_SEC: u64 = 10;
pub const MIN_RETRY_SEC: u64 = 5;
pub const MAX_ADJ_RETRY_SEC: u64 = 60;

/// [`Relay`](super::Relay) options
#[derive(Debug, Clone)]
pub struct RelayOptions {
    /// Proxy
    #[cfg(not(target_arch = "wasm32"))]
    pub proxy: Option<SocketAddr>,
    /// Allow/disallow read actions (default: true)
    read: Arc<AtomicBool>,
    /// Allow/disallow write actions (default: true)
    write: Arc<AtomicBool>,
    /// Enable/disable auto reconnection (default: true)
    reconnect: Arc<AtomicBool>,
    /// Retry connection time (default: 10 sec)
    ///
    /// Are allowed values `>=` 5 secs
    retry_sec: Arc<AtomicU64>,
    /// Automatically adjust retry seconds based on success/attempts (default: true)
    adjust_retry_sec: Arc<AtomicBool>,
}

impl Default for RelayOptions {
    fn default() -> Self {
        Self {
            #[cfg(not(target_arch = "wasm32"))]
            proxy: None,
            read: Arc::new(AtomicBool::new(true)),
            write: Arc::new(AtomicBool::new(true)),
            reconnect: Arc::new(AtomicBool::new(true)),
            retry_sec: Arc::new(AtomicU64::new(DEFAULT_RETRY_SEC)),
            adjust_retry_sec: Arc::new(AtomicBool::new(true)),
        }
    }
}

impl RelayOptions {
    /// New [`RelayOptions`]
    pub fn new() -> Self {
        Self::default()
    }

    /// Set proxy
    #[cfg(not(target_arch = "wasm32"))]
    pub fn proxy(mut self, proxy: Option<SocketAddr>) -> Self {
        self.proxy = proxy;
        self
    }

    /// Set read option
    pub fn read(self, read: bool) -> Self {
        Self {
            read: Arc::new(AtomicBool::new(read)),
            ..self
        }
    }

    pub(crate) fn get_read(&self) -> bool {
        self.read.load(Ordering::SeqCst)
    }

    /// Update read option
    pub fn update_read(&self, read: bool) {
        let _ = self
            .read
            .fetch_update(Ordering::SeqCst, Ordering::SeqCst, |_| Some(read));
    }

    /// Set write option
    pub fn write(self, write: bool) -> Self {
        Self {
            write: Arc::new(AtomicBool::new(write)),
            ..self
        }
    }

    pub(crate) fn get_write(&self) -> bool {
        self.write.load(Ordering::SeqCst)
    }

    /// Update write option
    pub fn update_write(&self, write: bool) {
        let _ = self
            .write
            .fetch_update(Ordering::SeqCst, Ordering::SeqCst, |_| Some(write));
    }

    /// Set reconnect option
    pub fn reconnect(self, reconnect: bool) -> Self {
        Self {
            reconnect: Arc::new(AtomicBool::new(reconnect)),
            ..self
        }
    }

    pub(crate) fn get_reconnect(&self) -> bool {
        self.reconnect.load(Ordering::SeqCst)
    }

    /// Set `reconnect` option
    pub fn update_reconnect(&self, reconnect: bool) {
        let _ = self
            .reconnect
            .fetch_update(Ordering::SeqCst, Ordering::SeqCst, |_| Some(reconnect));
    }

    /// Set retry seconds option
    pub fn retry_sec(self, retry_sec: u64) -> Self {
        let retry_sec = if retry_sec >= MIN_RETRY_SEC {
            retry_sec
        } else {
            DEFAULT_RETRY_SEC
        };
        Self {
            retry_sec: Arc::new(AtomicU64::new(retry_sec)),
            ..self
        }
    }

    pub(crate) fn get_retry_sec(&self) -> u64 {
        self.retry_sec.load(Ordering::SeqCst)
    }

    /// Set retry_sec option
    pub fn update_retry_sec(&self, retry_sec: u64) {
        if retry_sec >= MIN_RETRY_SEC {
            let _ = self
                .retry_sec
                .fetch_update(Ordering::SeqCst, Ordering::SeqCst, |_| Some(retry_sec));
        } else {
            tracing::warn!("Relay options: retry_sec it's less then the minimum value allowed (min: {MIN_RETRY_SEC} secs)");
        }
    }

    /// Automatically adjust retry seconds based on success/attempts (default: true)
    pub fn adjust_retry_sec(self, adjust_retry_sec: bool) -> Self {
        Self {
            adjust_retry_sec: Arc::new(AtomicBool::new(adjust_retry_sec)),
            ..self
        }
    }

    pub(crate) fn get_adjust_retry_sec(&self) -> bool {
        self.adjust_retry_sec.load(Ordering::SeqCst)
    }

    /// Set adjust_retry_sec option
    pub fn update_adjust_retry_sec(&self, adjust_retry_sec: bool) {
        let _ = self
            .adjust_retry_sec
            .fetch_update(Ordering::SeqCst, Ordering::SeqCst, |_| {
                Some(adjust_retry_sec)
            });
    }
}

/// [`Relay`](super::Relay) send options
#[derive(Debug, Clone, Copy)]
pub struct RelaySendOptions {
    /// Skip wait for disconnected relay (default: true)
    pub skip_disconnected: bool,
    /// Timeout for sending event (default: 10 secs)
    pub timeout: Duration,
}

impl Default for RelaySendOptions {
    fn default() -> Self {
        Self {
            skip_disconnected: true,
            timeout: DEFAULT_SEND_TIMEOUT,
        }
    }
}

impl RelaySendOptions {
    /// New default [`RelaySendOptions`]
    pub fn new() -> Self {
        Self::default()
    }

    /// Skip wait for disconnected relay (default: true)
    pub fn skip_disconnected(self, value: bool) -> Self {
        Self {
            skip_disconnected: value,
            ..self
        }
    }

    /// Timeout for sending event (default: 10 secs)
    ///
    /// If `None`, the default timeout will be used
    pub fn timeout(self, value: Option<Duration>) -> Self {
        Self {
            timeout: value.unwrap_or(DEFAULT_SEND_TIMEOUT),
            ..self
        }
    }
}

/// Filter options
#[derive(Debug, Clone, Copy, Default)]
pub enum FilterOptions {
    /// Exit on EOSE
    #[default]
    ExitOnEOSE,
    /// After EOSE is received, keep listening for N more events that match the filter, then return
    WaitForEventsAfterEOSE(u16),
    /// After EOSE is received, keep listening for matching events for [`Duration`] more time, then return
    WaitDurationAfterEOSE(Duration),
}

/// Relay Pool Options
#[derive(Debug, Clone, Copy)]
pub struct RelayPoolOptions {
    /// Notification channel size (default: 1024)
    pub notification_channel_size: usize,
    /// Task channel size (default: 1024)
    pub task_channel_size: usize,
    /// Shutdown on [RelayPool](super::pool::RelayPool) drop
    pub shutdown_on_drop: bool,
}

impl Default for RelayPoolOptions {
    fn default() -> Self {
        Self {
            notification_channel_size: 1024,
            task_channel_size: 1024,
            shutdown_on_drop: false,
        }
    }
}

impl RelayPoolOptions {
    /// New default options
    pub fn new() -> Self {
        Self::default()
    }

    /// Shutdown on [`RelayPool`](super::pool::RelayPool) drop
    pub fn shutdown_on_drop(self, value: bool) -> Self {
        Self {
            shutdown_on_drop: value,
            ..self
        }
    }
}

/// Negentropy reconciliation options
#[derive(Debug, Clone, Copy)]
pub struct NegentropyOptions {
    /// Timeout to check if negentropy it's supported (default: 10 secs)
    pub initial_timeout: Duration,
    /* /// Timeout for messages from relay (default: 10)
    ///
    /// If relay reply only to first messages, this timeout will stop the loop.
    pub recv_timeout: Duration, */
    /// Static timeout to get needed events from relay (default: 10)
    ///
    /// This timeout is added to `relative_get_events_timeout`
    pub static_get_events_timeout: Duration,
    /// Relative timeout to get needed events from relay (default: 250 ms per event)
    ///
    /// This timeout is added to `static_get_events_timeout`
    pub relative_get_events_timeout: Duration,
    /// Timeout for sending events to relay (default: 30 secs)
    pub batch_send_timeout: Duration,
    /// Bidirectional Sync (default: false)
    ///
    /// If `true`, perform the set reconciliation on each side.
    pub bidirectional: bool,
}

impl Default for NegentropyOptions {
    fn default() -> Self {
        Self {
            initial_timeout: Duration::from_secs(10),
            //recv_timeout: Duration::from_secs(600),
            static_get_events_timeout: Duration::from_secs(10),
            relative_get_events_timeout: Duration::from_millis(250),
            batch_send_timeout: Duration::from_secs(30),
            bidirectional: false,
        }
    }
}

impl NegentropyOptions {
    /// New default [`NegentropyOptions`]
    pub fn new() -> Self {
        Self::default()
    }

    /// Timeout to check if negentropy it's supported (default: 10 secs)
    pub fn initial_timeout(mut self, initial_timeout: Duration) -> Self {
        self.initial_timeout = initial_timeout;
        self
    }

    /* /// Timeout for messages from relay (default: 10)
    ///
    /// If relay reply only to first messages, this timeout will stop the loop.
    pub fn recv_timeout(mut self, recv_timeout: Duration) -> Self {
        self.recv_timeout = recv_timeout;
        self
    } */

    /// Static timeout to get needed events from relay (default: 10)
    ///
    /// This timeout is added to `relative_get_events_timeout`
    pub fn static_get_events_timeout(mut self, static_get_events_timeout: Duration) -> Self {
        self.static_get_events_timeout = static_get_events_timeout;
        self
    }

    /// Relative timeout to get needed events from relay (default: 250 ms per event)
    ///
    /// This timeout is added to `static_get_events_timeout`
    pub fn relative_get_events_timeout(mut self, relative_get_events_timeout: Duration) -> Self {
        self.relative_get_events_timeout = relative_get_events_timeout;
        self
    }

    /// Timeout for sending events to relay (default: 30 secs)
    pub fn batch_send_timeout(mut self, batch_send_timeout: Duration) -> Self {
        self.batch_send_timeout = batch_send_timeout;
        self
    }

    /// Bidirectional Sync (default: false)
    ///
    /// If `true`, perform the set reconciliation on each side.
    pub fn bidirectional(mut self, bidirectional: bool) -> Self {
        self.bidirectional = bidirectional;
        self
    }
}
